using Godot;
using System;

public enum ActorType 
{
	Player,
	Block, 
	Bgo,
	Npc,
	Warp,
	Liquid
	
}

[GlobalClass, Tool]
public abstract partial class Actor : CharacterBody2D
{
	private float            Timer      { get; set; } =0f;     
	private Sprite2D 		 Sprite 	{ get; set; }
	private CollisionShape2D Collider 	{ get; set; }

	

    float _imageTimer;
	uint  _imageFrame;
	uint  _imageFrames;

	[ExportGroup("Images")]
	[Export] public Texture2D    Image 			{ get; set; }
	[Export] public  int         ImageSpeed 	{ get; set; } = 8;
	[Export] public uint         ImageFrame 	
	{
		get  => Math.Clamp(_imageFrame, 0, GetFrames());
		set 
		{
			if (_imageFrame != value)
			{
				_imageFrame  = (uint)Mathf.Wrap(value, 0, Mathf.Max(0, GetFrames()));
			}
		}
	}
	[Export] public uint         ImageFrames 	
	{
		get  => Math.Clamp(_imageFrames, 0, GetFramesX() * GetFramesY());
		set 
		{
			if (_imageFrames != value)
			{
				_imageFrames  = Math.Clamp(value, 0, GetFramesX() * GetFramesY());
			}
		}
	}
	[Export] public uint         ImageOffset 	{ get; set; } = 0;

	[ExportGroup("Bounds")]
	[Export] public Rect2I       Bounds   			 { get; set; } = new Rect2I
	(
		0,
		0,
		16,
		16
	);
	[ExportGroup("Bounds/Overrides")]
	[Export] public Rect2I       GfxBoundsOverride   { get; set; }
	[Export] public Rect2I       HitBoundsOverride   { get; set; }
	[ExportGroup("Collision/Npc")]
	[Export] public bool IsNpcHurtLeft 			{ get; set; }
	[Export] public bool IsNpcHurtTop 			{ get; set; }
	[Export] public bool IsNpcHurtRight  		{ get; set; }
	[Export] public bool IsNpcHurtBottom 		{ get; set; }
	[Export] public bool IsNpcBlockLeft 		{ get; set; }
	[Export] public bool IsNpcBlockTop 			{ get; set; }
	[Export] public bool IsNpcBlockRight  		{ get; set; }
	[Export] public bool IsNpcBlockBottom 		{ get; set; }
	[ExportGroup("Collision/Player")]
	[Export] public bool IsPlayerHurtLeft 		{ get; set; }
	[Export] public bool IsPlayerHurtTop 		{ get; set; }
	[Export] public bool IsPlayerHurtRight  	{ get; set; }
	[Export] public bool IsPlayerHurtBottom 	{ get; set; }
	[Export] public bool IsPlayerBlockLeft 		{ get; set; }
	[Export] public bool IsPlayerBlockTop 		{ get; set; }
	[Export] public bool IsPlayerBlockRight  	{ get; set; }
	[Export] public bool IsPlayerBlockBottom 	{ get; set; }

    public override void _Ready  ()
    {
        base._Ready();
		ReadySprite  ();
		ReadyCollider();
	}
    public override void _Process(double delta)
    {
        base._Process(delta);
		ProcessSprite  ();
		ProcessCollider();
    }


	private void    ReadySprite()
	{
		Sprite?.QueueFree();
		Sprite   = new Sprite2D();
		AddChild(Sprite  , true, InternalMode.Front);
	}
	private void    ReadyCollider()
	{
		Collider?.QueueFree();
        Collider = new CollisionShape2D
        {
            Shape = new RectangleShape2D()
        };
		AddChild(Collider, true, InternalMode.Front);
	}
	private void  	ProcessSprite ()
	{
		_imageTimer 	   -= (float)(GetProcessDeltaTime() * 60);
		if (_imageTimer <= 0)
		{
			_imageTimer = ImageSpeed;
			var dir = (int)Mathf.Sign(ImageSpeed);
			var min = 0;
			var max = ImageFrames;
			if (ImageSpeed != 0)
			{
				ImageFrame = (uint)(((int)ImageFrame) + dir);

				while(dir > 0 && ImageFrame > max)
					ImageFrame -= ImageFrames;
				while(dir < 0 && ImageFrame < min)
					ImageFrame += ImageFrames;
			}
		}

		if (Sprite == null) return;
		Sprite.Texture  = Image;
		Sprite.Frame    = (int)ImageFrame;
		Sprite.Hframes  = (int)GetFramesX();
		Sprite.Vframes  = (int)GetFramesY();
		Sprite.Position = GetGfxBounds().Position;
	}
	private void    ProcessCollider()
	{
		if (Collider == null) return;
		
		Collider.Position = GetHitBounds().Position;
		Collider.Visible  = false;
		if (Collider.Shape is RectangleShape2D shape)
			shape.Size    = GetHitBounds().Size;

		if (TestMoveLeft  ())
		{
			if (GetActorType() == ActorType.Player)
			{
				if (IsPlayerBlockLeft)
                	OnPlayerBlockLeft();
			
				if (IsPlayerHurtLeft)
                	OnPlayerHurtLeft();
			}
			if (GetActorType() == ActorType.Npc)
			{
				if (IsNpcBlockLeft)
                	OnNpcBlockLeft();
			
				if (IsNpcHurtLeft)
                	OnNpcHurtLeft();
			}
		}
		if (TestMoveTop   ())
		{
			if (GetActorType() == ActorType.Player)
			{
				if (IsPlayerBlockTop)
					OnPlayerBlockTop();
			
				if (IsPlayerHurtTop)
					OnPlayerHurtTop();
			}
			if (GetActorType() == ActorType.Npc)
			{
				if (IsNpcBlockTop)
					OnNpcBlockTop();
			
				if (IsNpcHurtTop)
					OnNpcHurtTop();				
			}
		}
		if (TestMoveRight ())
		{
			if (GetActorType() == ActorType.Player)
			{
				if (IsPlayerBlockRight)
					OnPlayerBlockRight();
			
				if (IsPlayerHurtRight)
					OnPlayerHurtRight();
			}
			if (GetActorType() == ActorType.Npc)
			{
				if (IsNpcBlockRight)
					OnNpcBlockRight();
			
				if (IsNpcHurtRight)
					OnNpcHurtRight();
			}
		}
		if (TestMoveBottom())
		{
			if (GetActorType() == ActorType.Player)
			{
				if (IsPlayerBlockBottom)
					OnPlayerBlockBottom();
				
				if (IsPlayerHurtBottom)
					OnPlayerHurtBottom();
			}
			if (GetActorType() == ActorType.Npc)
			{
				if (IsNpcBlockBottom)
					OnNpcBlockBottom();
				
				if (IsNpcHurtBottom)
					OnNpcHurtBottom();
			}
		}
	}


    protected virtual void OnNpcBlockLeft()
    {
        
    }
    protected virtual void OnNpcHurtLeft()
    {
        
	}
    protected virtual void OnNpcBlockTop()
    {
        
    }
    protected virtual void OnNpcHurtTop()
    {
        
    }
    protected virtual void OnNpcBlockRight()
    {
        
    }
    protected virtual void OnNpcHurtRight()
    {
        
    }
    protected virtual void OnNpcBlockBottom()
    {
        
    }
    protected virtual void OnNpcHurtBottom()
    {
        
    }

    protected virtual void OnPlayerBlockLeft()
    {
        
    }
    protected virtual void OnPlayerHurtLeft()
    {
        
	}
    protected virtual void OnPlayerBlockTop()
    {
        
    }
    protected virtual void OnPlayerHurtTop()
    {
        
    }
    protected virtual void OnPlayerBlockRight()
    {
        
    }
    protected virtual void OnPlayerHurtRight()
    {
        
    }
    protected virtual void OnPlayerBlockBottom()
    {
        
    }
    protected virtual void OnPlayerHurtBottom()
    {
        
    }


    public Rect2I 	GetGfxBounds()
	{
		var bounds = Bounds;
		if (GfxBoundsOverride.Position.X != 0)
			bounds = new Rect2I(new(GfxBoundsOverride.Position.X, bounds.Position.Y), bounds.Size);
		if (GfxBoundsOverride.Position.Y != 0)
			bounds = new Rect2I(new(bounds.Position.X, GfxBoundsOverride.Position.Y), bounds.Size);
			
		if (GfxBoundsOverride.Size.X     != 0)
			bounds = new Rect2I(bounds.Position, new(GfxBoundsOverride.Size.X, bounds.Size.Y));
		if (GfxBoundsOverride.Size.Y     != 0)
			bounds = new Rect2I(bounds.Position, new(bounds.Size.X, GfxBoundsOverride.Size.Y));
		return bounds;
	}
	public Rect2I 	GetHitBounds()
	{
		var bounds = Bounds;
		if (HitBoundsOverride.Position.X != 0)
			bounds = new Rect2I(new(HitBoundsOverride.Position.X, bounds.Position.Y), bounds.Size);
		if (HitBoundsOverride.Position.Y != 0)
			bounds = new Rect2I(new(bounds.Position.X, HitBoundsOverride.Position.Y), bounds.Size);
			
		if (HitBoundsOverride.Size.X     != 0)
			bounds = new Rect2I(bounds.Position, new(HitBoundsOverride.Size.X, bounds.Size.Y));
		if (HitBoundsOverride.Size.Y     != 0)
			bounds = new Rect2I(bounds.Position, new(bounds.Size.X, HitBoundsOverride.Size.Y));
		return bounds;
	}

	public uint GetFrames ()
	{
		return (uint)Mathf.Min(ImageFrames, GetFramesX() * GetFramesY());
	}
	public uint GetFramesX()
	{
		if (Image == null) return 0;
		return (uint)(Image.GetSize().X / GetGfxBounds().Size.X);
	}
	public uint GetFramesY()
	{
		if (Image == null) return 0;
		return (uint)(Image.GetSize().Y / GetGfxBounds().Size.Y);
	}

	public bool TestMove      ()
	{
		return TestMoveLeft() || TestMoveTop() || TestMoveRight() || TestMoveBottom();
	}
	public bool TestMoveLeft  ()
	{
		return Velocity.X > 0 && TestMove(Transform, Velocity);
	}
	public bool TestMoveTop   ()
	{
		return Velocity.Y > 0 && TestMove(Transform, Velocity);
	}
	public bool TestMoveRight ()
	{
		return Velocity.X < 0 && TestMove(Transform, Velocity);
	}
	public bool TestMoveBottom()
	{
		return Velocity.Y < 0 && TestMove(Transform, Velocity);
	}

	public abstract ActorType GetActorType();
}
